from dataclasses import dataclass
from typing import Any

from dataclasses_json import dataclass_json
from pydantic import BaseModel


@dataclass_json
@dataclass
class EntityLabel:
    label: str
    name: str
    description: str | None


@dataclass_json
@dataclass
class EntityReference:
    start: int
    end: int
    label: str
    text: str | None = None
    reason: str | None = None


@dataclass_json
@dataclass
class TextwithEntities:
    text: str
    ents: list[EntityReference]


class ChatMessage(BaseModel):
    role: str
    content: str
    name: str | None = None


class FunctionParameter(BaseModel):
    type: str
    description: str | None


class FunctionArrayParameter(BaseModel):
    type: str
    description: str | None
    items: Any | None


class FunctionParameters(BaseModel):
    type: str
    properties: dict[str, FunctionParameter]
    required: list[str]


class ChatFunction(BaseModel):
    name: str
    description: str
    dependencies: list[str] | None
    parameters: FunctionParameters



def deep_merge(d1: dict[str, any], d2: dict[str, any]):
    for k, v in d2.items():
        if k not in d1 or d1[k] is None:
            d1[k] = v
        elif isinstance(d1[k], dict) and isinstance(v, dict):
            deep_merge(d1[k], v)
        else:
            d1[k] += str(v)



import json


def extract_partial_list(s, key="entities"):
    if not isinstance(s, str) or s == "":
        return []

    # remove everything before the first occurrence of the key
    field = f'"{key}":'
    s = field.join(s.split(field)[1:]).strip()
    if len(s) == 0:
        return []

    if not s.startswith("["):
        return []

    if s.endswith("},"):
        s = s[:-1]

    if not s.rstrip().endswith("}"):
        return []

    s += "]"
    try:
        return json.loads(s)
    except json.decoder.JSONDecodeError:
        return []



"""This module contains the entity recognition pipeines
We can rely on the spacy library for this task as well
as their spacy-llm library for the language model.

We will be providing a custom implementation of spacy-llm
in order to support the streaming responses from the API
and will still make the results play well with spacy to enable
custom training and other features.

Ideally ER should be done locally with a transformer-style
model trained on a given domain.  Until we get there,
we will use LLMs to label the data.

Base case:
 * Run entity recognition on the text, stream back spacy-compatible entity spans

Nice to have
 * Run ER multiple times with noise and only choose a subset of labels
 * Aggregate the highest confidence common labels

Ideal case:
 * For labels with high disagreement, have a human label the data
 * Create a gold-star dataset for training a custom model
 * Train a custom model on the data
 * Evaluate the performance of the model and compare to the LLM in performance, cost and time

"""


import contextlib
import json
import re

from promptedgraphs.config import Config
from promptedgraphs.llms.openai_streaming import (
    GPT_MODEL,
    streaming_chat_completion_request,
)
from promptedgraphs.models import ChatMessage, EntityReference
from promptedgraphs.parsers import extract_partial_list

# Name and description of entity types


SYSTEM_MESSAGE = """
You are an expert Named Entity Recognition (NER) system.
Your task is to accept Text as input and extract named entities to be passed to the `{name}` function.

Entities must have one of the following labels: {label_list}.
If a span is not an entity label it: `==NONE==`.

{label_definitions}
"""


def build_message_list(
    text: str,
    label_list: list[str],
    labels: dict[str, str],
    name="extract_entities",
):
    label_definitions = """Below are definitions of each label to help aid you in what kinds of named entities to extract for each label.
Assume these definitions are written by an expert and follow them closely.\n\n""" + "\n".join(
        [f" * {label}: {labels[label]}" for label in label_list]
    )

    return [
        ChatMessage(
            role="system",
            content=SYSTEM_MESSAGE.format(
                name=name, label_list=label_list, label_definitions=label_definitions
            ),
        ),
        ChatMessage(role="user", content=text),
    ]


def build_function_spec(
    label_list: list[str],
    function_name="extract_entities",
    key="entities",
    description="",
    include_reason=True,
):
    spec = [
        {
            "name": function_name,
            "type": "function",
            "description": description,
            "parameters": {
                "type": "object",
                "properties": {
                    key: {
                        "type": "array",
                        "items": {
                            "type": "object",
                            "description": "The raw text and label for each entity occurrence",
                            "properties": {
                                "text_span": {
                                    "type": "string",
                                    "description": "The exact text of referenced entity.",
                                },
                                "is_entity": {
                                    "type": "boolean",
                                    "description": "A boolean indicating if the span is an entity label.",
                                },
                                "label": {
                                    "type": "string",
                                    "enum": label_list + ["==NONE=="],
                                    "description": "The label of the entity.",
                                },
                                "reason": {
                                    "type": "string",
                                    "description": "A short description of why that label was selected.",
                                },
                            },
                            "required": ["text_span", "is_entity", "label", "reason"],
                        },
                    }
                },
            },
        }
    ]
    if not include_reason:
        spec[0]["parameters"]["properties"][key]["items"]["properties"].pop("reason")
        spec[0]["parameters"]["properties"][key]["items"]["required"] = [
            "text_span",
            "is_entity",
            "label",
        ]
    return spec


def _format_entities(s, text):
    for entity in s:
        if entity["is_entity"]:
            for match in re.finditer(entity["text_span"], text):
                yield EntityReference(
                    start=match.start(),
                    end=match.end(),
                    text=entity["text_span"],
                    label=entity["label"],
                    reason=entity.get("reason"),
                )


async def extract_entities(
    text: str,
    labels: dict[str, str],
    config: Config,
    name="entities",
    description: str | None = "Extract Entities from text",
    include_reason=True,
    model=GPT_MODEL,
    temperature=0.2,
):
    label_list = sorted(labels.keys())
    messages = build_message_list(
        text,
        label_list,
        labels=labels,
        name=f"extract_{name}".lower().replace(" ", "_"),
    )

    functions = build_function_spec(
        label_list,
        function_name=f"extract_{name}".lower().replace(" ", "_"),
        key=name,
        description=description,
        include_reason=include_reason,
    )

    count = 0
    payload = ""
    async for msg in streaming_chat_completion_request(
        messages=messages,
        functions=functions,
        model=model,
        config=config,
        temperature=temperature,
    ):
        if msg.data is None or msg.data == "":
            continue

        if msg.data == "[DONE]":
            with contextlib.suppress(json.decoder.JSONDecodeError):
                s = json.loads(payload).get(name, [])
                for entity in _format_entities(s[count:], text):
                    yield entity

            break

        # TODO try catch for malformed json
        data = json.loads(msg.data)

        choices = data.get("choices")
        if choices is None:
            continue

        delta = choices[0].get("delta")

        # TODO rewrite this to be more robust streaming json parser
        payload += delta.get("function_call", {}).get("arguments", "")
        s = extract_partial_list(payload, key=name)
        if s is None or len(s) == 0 or len(s) <= count:
            continue

        for entity in _format_entities(s[count:], text):
            yield entity
        count = len(s)



import pyfiglet
from rich import print
from typer import Typer

from promptedgraphs import __description__ as DESCRIPTION
from promptedgraphs import __title__ as NAME
from promptedgraphs import __version__ as VERSION
from promptedgraphs.config import load_config


def banner():
    return pyfiglet.figlet_format(NAME.replace("_", " ").title(), font="slant").rstrip()


app = Typer(help=f"{(NAME or '').replace('_', ' ').title()} CLI")


@app.command()
def info():
    """Prints info about the package"""
    print(f"{banner()}\n")
    print(f"{NAME}: {DESCRIPTION}")
    print(f"Version: {VERSION}\n")
    print(load_config())


@app.command()
def main():
    """Main Function"""
    print(f"{banner()}\n")
    print(
        "This is your default command-line interface.  Feel free to customize it as you see fit.\n"
    )



"""The purpose of this class is to extract data from text
and return it as a structured object or a list of structured objects.

Base case:
 * Provide text and specify the return type as a Pydantic model and it returns an instance of that model

"""


import contextlib
import json
import re

from pydantic import BaseModel

from promptedgraphs.config import Config
from promptedgraphs.llms.openai_streaming import (
    GPT_MODEL,
    streaming_chat_completion_request,
)
from promptedgraphs.models import ChatMessage
from promptedgraphs.parsers import extract_partial_list

SYSTEM_MESSAGE = """
You are a Qualitative User Researcher and Linguist. Your task is to extract structured data from text to be passed into python's `{name}(BaseModel)` pydantic class.
Maintain as much verbatim text as possible, light edits are allowed, feel free to remove any text that is not relevant to the label.
If there is not information applicable for a particular field, use the value `==NA==`.

In particular look for the following fields: {label_list}.

{label_definitions}
"""


def add_space_before_capital(text):
    return re.sub(r"(?<=[a-z])(?=[A-Z])", " ", text)


def camelcase_to_words(text):
    if len(text) <= 1:
        return text.replace("_", " ")
    if text.lower() == text:
        return text.replace("_", " ")
    return add_space_before_capital(text[0].upper() + text[1:]).replace("_", " ")


def format_fieldinfo(key, v: dict):
    l = f"`{key}`: {v.get('title','')} - {v.get('description','')}".strip()
    if v.get("annotation"):
        annotation = v.get("annotation")
        if annotation == "str":
            annotation = "string"
        l += f"\n\ttype: {annotation}"
    if v.get("examples"):
        l += f"\n\texamples: {v.get('examples')}"
    return l.rstrip()


def create_messages(text, name, labels):
    label_list = list(labels.keys())

    label_definitions = """Below are definitions of each field to help aid you in what kinds of structured data to extract for each label.
Assume these definitions are written by an expert and follow them closely.\n\n""" + "\n".join(
        [f" * {format_fieldinfo(k, dict(labels[k]))}" for k in label_list]
    )

    messages = [
        ChatMessage(
            role="system",
            content=SYSTEM_MESSAGE.format(
                name=name, label_list=label_list, label_definitions=label_definitions
            ),
        ),
        ChatMessage(role="user", content=text),
    ]

    return messages


def create_functions(is_parent_list, schema, name, description, labels, fn_name):
    if is_parent_list:
        return [
            {
                "name": name,
                "type": "function",
                "description": "Extract a non-empty list of structured data from text",
                "parameters": {
                    "type": "object",
                    "properties": {
                        fn_name: {
                            "type": "array",
                            "items": {
                                "type": "object",
                                "description": description,
                                "properties": {
                                    k: {
                                        "type": v.get("type", "string"),
                                    }
                                    for k, v in labels.items()
                                },
                                "required": schema.get("required", []),
                            },
                        }
                    },
                },
            }
        ]

    return [
        {
            "name": name,
            "type": "function",
            "description": description,
            "parameters": {
                "type": "object",
                "properties": {
                    k: {
                        "type": v.get("type", "string"),
                    }
                    for k, v in labels.items()
                },
                "required": schema.get("required", []),
            },
        }
    ]


async def extract_data(
    text: str,
    output_type: list[BaseModel] | BaseModel,
    config: Config,
    model=GPT_MODEL,
    temperature=0.2,
):
    if is_parent_list := str(output_type).lower().startswith("list"):
        assert (
            len(output_type.__args__) == 1
        ), "Only one type argument is allowed for list types"
        output_type = output_type.__args__[0]
    assert issubclass(output_type, BaseModel), "output_type must be a Pydantic model"

    schema = output_type.model_json_schema()
    name = schema["title"]
    description = schema["description"]
    labels = schema["properties"]
    fn_name = camelcase_to_words(name).replace(" ", "_").lower()

    messages = create_messages(text, name, labels)
    functions = create_functions(
        is_parent_list, schema, name, description, labels, fn_name
    )

    count = 0
    payload = ""
    async for msg in streaming_chat_completion_request(
        messages=messages,
        functions=functions,
        model=model,
        config=config,
        temperature=temperature,
    ):
        if msg.data is None or msg.data == "":
            continue

        if msg.data == "[DONE]":
            with contextlib.suppress(json.decoder.JSONDecodeError):
                if is_parent_list:
                    s = json.loads(payload).get(name, [])
                    for data in s[count:]:
                        yield output_type(**data)
                else:
                    data = json.loads(payload)
                    yield output_type(**data)
            break

        # TODO try catch for malformed json
        if msg.event == "error":
            print("ERROR", msg.data)
            break

        data = json.loads(msg.data)

        choices = data.get("choices")
        if choices is None:
            continue

        delta = choices[0].get("delta")

        # TODO rewrite this to be more robust streaming json parser
        payload += delta.get("function_call", {}).get("arguments", "")

        if is_parent_list:
            s = extract_partial_list(payload, key=fn_name)
            if s is None or len(s) == 0 or len(s) <= count:
                continue

            for data in s[count:]:
                yield output_type(**data)

            count = len(s)



import json

import tiktoken


def num_tokens_from_messages(messages, model="gpt-3.5-turbo-0613"):
    """Returns the number of tokens used by a list of messages.
    See https://github.com/openai/openai-python/blob/main/chatml.md for information on how messages are converted to tokens.
    """
    try:
        encoding = tiktoken.encoding_for_model(model)
    except KeyError:
        encoding = tiktoken.get_encoding("cl100k_base")
    if model == "gpt-3.5-turbo-0613":
        num_tokens = 0
        for message in messages:
            num_tokens += (
                4  # every message follows <im_start>{role/name}\n{content}<im_end>\n
            )
            for key, value in message.items():
                num_tokens += len(encoding.encode(value))
                if key == "name":  # if there's a name, the role is omitted
                    num_tokens += -1  # role is always required and always 1 token
        num_tokens += 2  # every reply is primed with <im_start>assistant
        return num_tokens
    else:
        encoding = tiktoken.encoding_for_model(model)
        return len(
            encoding.encode(
                "<|endoftext|>".join(
                    [l for m in messages for l in m.values() if l is not None]
                ),
                allowed_special=set(encoding._special_tokens.keys()),
            )
        )


def estimate_tokens(json_data, model="gpt-3.5-turbo-0613"):
    # Adjust max_tokens based on message length and function length
    message_tokens = num_tokens_from_messages(
        json_data.get("messages", []), model=model
    )

    encoding = tiktoken.encoding_for_model(model)
    function_tokens = len(
        encoding.encode(
            "<|endoftext|>".join(
                [
                    json.dumps(l)
                    for m in json_data.get("functions", [])
                    for l in m.values()
                    if l is not None
                ]
            ),
            allowed_special=set(encoding._special_tokens.keys()),
        )
    )
    return function_tokens + message_tokens



"""Loads the configuration file for the QuantReady package."""
# Load the configuration file
import os
from dataclasses import dataclass, field
from pathlib import Path

from dotenv import load_dotenv

from promptedgraphs import __description__ as description
from promptedgraphs import __title__ as name
from promptedgraphs import __version__ as version


@dataclass
class Config:
    """Configuration class for PromptedGraphs"""

    name: str = name
    description: str = description
    version: str = version
    openai_api_key: str | None = field(
        default_factory=lambda: os.getenv("OPENAI_API_KEY")
    )


def load_config() -> Config:
    load_dotenv(
        verbose=True,
        dotenv_path=Path(__file__).parent.joinpath(".env"),
        override=False,
    )
    return Config(
        name=name,
        description=description,
        version=version,
    )



# https://github.com/openai/openai-cookbook/blob/60b12dfad1b6e7b32c4a6f1edff3b94c946b467d/examples/How_to_call_functions_with_chat_models.ipynb
from collections.abc import AsyncGenerator

from httpx import AsyncClient, ReadTimeout
from sse_starlette import ServerSentEvent

from promptedgraphs.config import Config
from promptedgraphs.llms.openai_token_counter import estimate_tokens
from promptedgraphs.models import ChatFunction, ChatMessage

GPT_MODEL = "gpt-3.5-turbo-0613"
GPT_MODEL_BIG_CONTEXT = "gpt-3.5-turbo-16k-0613"


# @retry(wait=wait_random_exponential(min=1, max=20), stop=stop_after_attempt(3))
async def streaming_chat_completion_request(
    messages: list[ChatMessage] | None,
    functions: list[ChatFunction] | None = None,
    model=GPT_MODEL,
    config: Config | None = None,
    temperature=0.2,
    max_tokens=4000,
) -> AsyncGenerator[bytes, None]:
    assert config and config.openai_api_key is not None, "OpenAI API Key not found"

    url = "https://api.openai.com/v1/chat/completions"
    headers = {
        "Content-Type": "application/json",
        "Authorization": f"Bearer {config.openai_api_key}",
    }

    json_data = {
        "model": model,
        "messages": [m.model_dump(exclude_none=True) for m in messages or []],
        "temperature": temperature,
        "max_tokens": max_tokens,
        "stream": True,
    }
    if functions is not None and len(functions) > 0:
        json_data["functions"] = [
            f if isinstance(f, dict) else f.model_dump(exclude_none=True)
            for f in functions
        ]

    token_count_approx = estimate_tokens(json_data, model=model)

    if token_count_approx >= 4_096:
        model = GPT_MODEL_BIG_CONTEXT
        json_data["max_tokens"] = 16_384
        json_data["model"] = model
        json_data["messages"][-1]["content"] = json_data["messages"][-1][
            "content"
        ].strip()[:40_000]

    json_data["max_tokens"] = min(
        max(json_data["max_tokens"] - int(token_count_approx), 200), 16_384
    )

    async with AsyncClient(timeout=None) as client:
        try:
            response = await client.post(url, headers=headers, json=json_data)
            if response.status_code != 200:
                raise ValueError(f"Failed to post to {url}. Response: {response.text}")

            async for chunk in response.aiter_lines():
                if chunk.startswith("data:"):
                    yield ServerSentEvent(data=chunk[5:].strip())
                elif chunk.startswith("event:"):
                    yield ServerSentEvent(event=chunk[6:].strip())
                elif chunk.startswith("id:"):
                    yield ServerSentEvent(id=chunk[3:].strip())
                elif chunk.startswith("retry:"):
                    yield ServerSentEvent(retry=chunk[6:].strip())
                else:
                    yield ServerSentEvent(data=chunk.strip())
        except GeneratorExit:
            pass  # Handle the generator being closed, if necessary
        except ReadTimeout:
            yield ServerSentEvent(
                data="timeout: Timeout reading the response", event="error"
            )
        except Exception as e:
            yield ServerSentEvent(data=str(e), event="error")
